from pyspark.sql import functions as F
from pyspark.sql.functions import udf
from pyspark.sql.types import DoubleType, LongType, StructField, IntegerType

import pandas as pd
from statsmodels.stats.weightstats import DescrStatsW


@udf(LongType())
def histogram_sum(values):
    """Returns the weighted sum of the histogram.

    Note that this is the sum *post*-quantization, which amounts to a
    left-hand-rule discrete integral of the histogram. It is therefore
    likely to be an underestimate of the true sum.
    """
    if values is None:
        return None
    return sum(int(k) * v for k, v in values.items())


@udf(DoubleType())
def histogram_mean(values):
    """Returns the mean of values in a histogram.

    This mean relies on the sum *post*-quantization, which amounts to a
    left-hand-rule discrete integral of the histogram. It is therefore
    likely to be an underestimate of the true mean.
    """
    if values is None:
        return None
    numerator = 0
    denominator = 0
    for k, v in values.items():
        numerator += int(k) * v
        denominator += v
    return numerator / float(denominator)


@udf(LongType())
def histogram_count(values):
    "Counts the number of events recorded by a histogram."
    if values is None:
        return None
    return sum(values.values())


def generate_quantile_udf(sdf, grouping_fields, bucket_field, count_field, quantiles):
    """Returns a UDF for computing histogram quantiles.

    Args:
        sdf: Spark DataFrame
        grouping_fields: list of strings representing columns from sdf
        bucket_field: Name of the field containing histogram buckets
        count_field: Name of the field containing counts
        quantiles: List of quantiles as floating point values on [0, 1]

    ``sdf`` and ``grouping_fields`` are necessary to infer the schema of the
    returned Spark DataFrame.

    For example, to compute per-user median ``FX_NEW_WINDOW_MS`` values for a recent
    sample of main_summary, you could write something like the example below::

        import pyspark.sql.functions as F
        ms = spark.table("main_summary")

        grouping_fields = ["client_id"]

        quantile_udf = generate_quantile_udf(
            sdf=ms,
            grouping_fields=grouping_fields,
            bucket_field="bucket",
            count_field="count",
            quantiles=[0.5, 0.95],
        )

        quantiles = (
            ms
            .filter(ms.submission_date_s3 == "20181001")
            .filter(ms.sample_id == "42")
            .select(
                (
                    F.explode(ms.histogram_content_fx_new_window_ms)
                    .alias("bucket", "count")
                ),
                *grouping_fields
            )
            .groupBy(*grouping_fields)
            .apply(quantile_udf)
            .toPandas()
        )

    Note that the ``quantile_udf`` returned by this function is used together
    with the ``.apply`` method of the grouped ``DataFrame``.

    This will return a table ``quantiles`` that looks like the table below.

    ========= ======== =====
    client_id quantile value
    ========= ======== =====
    id_1      0.5      2
    id_1      0.95     23
    id_2      0.5      4
    ...       ...      ...
    ========= ======== =====
    """
    schema = (
        sdf
        .select(*grouping_fields)
        .schema
        .add(StructField("quantile", DoubleType(), False))
        .add(StructField("value", DoubleType(), True))
    )

    @F.pandas_udf(schema, F.PandasUDFType.GROUPED_MAP)
    def udf(df):
        stats = DescrStatsW(
            df[bucket_field].astype(float), df[count_field]
        ).quantile(quantiles)
        stats = stats.rename_axis("quantile").reset_index(name="value").assign(_dummy=1)
        # Join the (constant) grouping variables back onto the result
        grouping_variables = df[grouping_fields].iloc[:1].assign(_dummy=1)
        result = grouping_variables.merge(stats, on="_dummy").drop("_dummy", axis=1)
        return result

    return udf


def generate_threshold_udf(sdf, grouping_fields, bucket_field, count_field, thresholds):
    """Returns a UDF for calculating the fraction of values in a histogram
    equal to or exceeding a threshold.

    Args:
        sdf: Spark DataFrame
        grouping_fields: list of strings representing columns from sdf
        bucket_field: Name of the field containing histogram buckets
        count_field: Name of the field containing counts
        thresholds: List of thresholds. Should probably align on the left edge
            of a bucket.

    `sdf` and `grouping_fields` are necessary to infer the schema of the
    returned Spark DataFrame.

    For example, to compute the per-user fractions of CONTENT_FRAME_TIME values
    equal to or exceeding 103% or 192% of a vsync from a sample of main_summary,
    you could write::

        import pyspark.sql.functions as F
        ms = spark.table("main_summary")

        grouping_fields = ["client_id"]

        threshold_udf = generate_threshold_udf(
            sdf=ms,
            grouping_fields=grouping_fields,
            bucket_field="bucket",
            count_field="count",
            thresholds=[103, 192],
        )

        thresholded = (
            ms
            .filter(ms.submission_date_s3 == "20181001")
            .filter(ms.sample_id == "42")
            .select(
                F.explode(ms.histogram_gpu_content_frame_time).alias("bucket", "count"),
                *grouping_fields
            )
            .groupBy(*grouping_fields)
            .apply(threshold_udf)
            .toPandas()
        )

    Note that the `threshold_udf` returned by this function is used together
    with the `.apply` method of the grouped DataFrame.

    This will return a table `thresholded` that looks like:

    =========  =========  ==================
    client_id  threshold  fraction_exceeding
    =========  =========  ==================
    id_1       103        0.994
    id_1       192        0.036
    id_2       103        0.978
    ...        ...        ...
    =========  =========  ==================
    """
    schema = (
        sdf
        .select(*grouping_fields)
        .schema
        .add(StructField("threshold", IntegerType(), False))
        .add(StructField("fraction_exceeding", DoubleType(), True))
    )

    @F.pandas_udf(schema, F.PandasUDFType.GROUPED_MAP)
    def udf(df):
        rows = []
        all_sum = float(df[count_field].sum())
        for tx in thresholds:
            fraction = (
                df.loc[df[bucket_field].astype(int) >= tx, count_field].sum() / all_sum
            )
            rows.append({"threshold": tx, "fraction_exceeding": fraction})
        data = pd.DataFrame(rows).assign(_dummy=1)
        grouping_variables = df[grouping_fields].iloc[:1].assign(_dummy=1)
        result = grouping_variables.merge(data, on="_dummy").drop("_dummy", axis=1)
        return result

    return udf



import pyspark.sql.functions as F
import pytest

from mozanalysis.experiment import Experiment
from mozanalysis.utils import add_days


def test_get_last_enrollment_date():
    exp = Experiment('a-stub', '20190101')
    exp_8d = Experiment('experiment-with-8-day-cohort', '20190101', 8)

    # When we have complete data for 20190114...
    the_fourteenth = '20190114'

    # ...We have 14 dates of data for those who enrolled on the 1st
    exp._get_last_enrollment_date(the_fourteenth, 14) == '20190101'

    # We don't have 14 dates of data for the 8-day cohort:
    with pytest.raises(ValueError):
        exp_8d._get_last_enrollment_date(the_fourteenth, 14)

    # We don't have 15 full dates of data for any users
    with pytest.raises(ValueError):
        exp._get_last_enrollment_date(the_fourteenth, 15)

    # And we certainly don't have 15 full dates for the 8-day cohort:
    with pytest.raises(ValueError):
        exp_8d._get_last_enrollment_date(the_fourteenth, 15)

    # For the 8-day cohort We have enough data for a 7 day window
    exp_8d._get_last_enrollment_date(the_fourteenth, 7) == '20190108'

    # Or a 2 day window
    exp_8d._get_last_enrollment_date(the_fourteenth, 2) == '20190108'

    # But not an 8 day window
    with pytest.raises(ValueError):
        exp_8d._get_last_enrollment_date(the_fourteenth, 8)

    # Of course the flexi-experiment has data for a 1 day window
    exp._get_last_enrollment_date(the_fourteenth, 1) == '20190114'


def test_get_last_data_date1():
    exp = Experiment('a-stub', '20190101')

    # When we have complete data for 20190114...
    the_fourteenth = '20190114'

    # When we don't specify num_days_enrollment we'll use all the data
    assert exp._get_last_data_date(the_fourteenth, 14) == the_fourteenth
    assert exp._get_last_data_date(the_fourteenth, 10) == the_fourteenth
    assert exp._get_last_data_date(the_fourteenth, 1) == the_fourteenth
    assert exp._get_last_data_date(the_fourteenth, 5) == the_fourteenth

    # But we don't have 15 full dates of data for any users
    with pytest.raises(ValueError):
        exp._get_last_data_date(the_fourteenth, 15)


def test_get_last_data_date2():
    exp_8d = Experiment('experiment-with-8-day-cohort', '20190101', 8)

    the_fourteenth = '20190114'

    with pytest.raises(ValueError):
        assert exp_8d._get_last_data_date(the_fourteenth, 14) == the_fourteenth

    with pytest.raises(ValueError):
        assert exp_8d._get_last_data_date(the_fourteenth, 10) == the_fourteenth

    # If we only need 1 date of data then the final enrollment is fixed:
    assert exp_8d._get_last_data_date(the_fourteenth, 1) == '20190108'
    assert exp_8d._get_last_data_date('20190131', 1) == '20190108'
    assert exp_8d._get_last_data_date('20201231', 1) == '20190108'

    # And it's fixed to the date of the final enrollment
    assert exp_8d._get_last_data_date(the_fourteenth, 1) \
        == '20190108' \
        == exp_8d._get_last_enrollment_date(the_fourteenth, 1)

    assert exp_8d._get_last_data_date(the_fourteenth, 5) == '20190112'


def _get_data_source(spark):
    clients_branches = [
        ('aaaa', 'control'),
        ('bbbb', 'test'),
    ]
    dates = [add_days('20181215', i) for i in range(32)]

    data_rows = [
        [
            client, submission_date_s3, {'a-stub': branch}, 1.
        ]
        for client, branch in clients_branches
        for submission_date_s3 in dates
    ]

    return spark.createDataFrame(
        data_rows,
        [
            "client_id",
            "submission_date_s3",
            "experiments",
            "constant_one",
        ],
    )


def _simple_return_agg_date(agg_fn, data_source):
    return data_source.select(
        agg_fn(data_source.submission_date_s3).alias('b')
    ).first()['b']


def test_filter_data_source_for_analysis_window(spark):
    start_date = '20190101'
    exp_8d = Experiment('experiment-with-8-day-cohort', start_date, 8)
    data_source = _get_data_source(spark)

    end_date = '20190114'

    # Are the fixtures sufficiently complicated that we're actually testing
    # things?
    assert _simple_return_agg_date(F.min, data_source) < start_date
    assert _simple_return_agg_date(F.max, data_source) > end_date

    filtered_ds = exp_8d.filter_data_source_for_analysis_window(
        data_source, end_date, 0, 3
    )

    assert _simple_return_agg_date(F.min, filtered_ds) == start_date
    assert _simple_return_agg_date(F.max, filtered_ds) == '20190110'

    filtered_ds_2 = exp_8d.filter_data_source_for_analysis_window(
        data_source, end_date, 2, 3
    )

    assert _simple_return_agg_date(F.min, filtered_ds_2) == add_days(start_date, 2)
    assert _simple_return_agg_date(F.max, filtered_ds_2) == '20190112'


def _get_enrollment_view(spark, slug):
    # `slug` is supplied so we reuse this fixture
    # with multiple slugs
    data_rows = [
        ['aaaa', slug, 'control', '20190101'],
        ['bbbb', slug, 'test', '20190101'],
        ['cccc', slug, 'control', '20190108'],
        ['dddd', slug, 'test', '20190109'],
        ['eeee', 'no', 'control', '20190101'],
    ]

    return spark.createDataFrame(
        data_rows,
        [
            "client_id",
            "experiment_slug",
            "branch",
            "enrollment_date",
        ],
    )


def test_filter_enrollments_for_analysis_window(spark, monkeypatch):
    exp = Experiment('a-stub', '20190101')
    _mock_exp(monkeypatch, exp)
    enrollments = exp.get_enrollments(spark)
    assert enrollments.count() == 4

    # With final data collected on '20190114', we have 7 dates of data
    # for 'cccc' enrolled on '20190108' but not for 'dddd' enrolled on
    # '20190109'.
    fe = exp.filter_enrollments_for_analysis_window(enrollments, '20190114', 7)
    assert fe.count() == 3


def _mock_exp(monkeypatch, exp):
    monkeypatch.setattr(
        exp, '_get_enrollments_view_normandy',
        lambda spark: _get_enrollment_view(spark, exp.experiment_slug)
    )
    monkeypatch.setattr(
        exp, '_get_enrollments_view_addon',
        lambda spark, addon_version: _get_enrollment_view(spark, exp.experiment_slug)
    )


def test_get_enrollments(spark, monkeypatch):
    # Experiment = _mock_exp(monkeypatch)

    exp = Experiment('a-stub', '20190101')
    _mock_exp(monkeypatch, exp)
    assert exp.get_enrollments(spark).count() == 4

    exp2 = Experiment('a-stub2', '20190102')
    _mock_exp(monkeypatch, exp2)
    enrl2 = exp2.get_enrollments(spark, study_type='addon')
    assert enrl2.count() == 2
    assert enrl2.select(F.min(enrl2.enrollment_date).alias('b')).first(
        )['b'] == '20190108'

    exp_8d = Experiment('experiment-with-8-day-cohort', '20190101', 8)
    _mock_exp(monkeypatch, exp_8d)
    enrl_8d = exp_8d.get_enrollments(spark)
    assert enrl_8d.count() == 3
    assert enrl_8d.select(F.max(enrl_8d.enrollment_date).alias('b')).first(
        )['b'] == '20190108'


def test_get_enrollments_debug_dupes(spark, monkeypatch):
    exp = Experiment('a-stub', '20190101')
    _mock_exp(monkeypatch, exp)

    enrl = exp.get_enrollments(spark)
    assert 'num_events' not in enrl.columns

    enrl2 = exp.get_enrollments(spark, debug_dupes=True)
    assert 'num_events' in enrl2.columns

    penrl2 = enrl2.toPandas()
    assert (penrl2['num_events'] == 1).all()


def test_get_per_client_data_doesnt_crash(spark):
    exp = Experiment('a-stub', '20190101', 8)
    enrollments = _get_enrollment_view(spark, exp.experiment_slug)
    data_source = _get_data_source(spark)

    exp.get_per_client_data(
        enrollments,
        data_source,
        [
            F.sum(data_source.constant_one).alias('something_meaningless'),
        ],
        '20190114',
        0,
        3
    )


def test_get_per_client_data_join(spark):
    exp = Experiment('a-stub', '20190101')

    enrollments = spark.createDataFrame(
        [
            ['aaaa', 'control', '20190101'],
            ['bbbb', 'test', '20190101'],
            ['cccc', 'control', '20190108'],
            ['dddd', 'test', '20190109'],
            ['annie-nodata', 'control', '20190101'],
            ['bob-badtiming', 'test', '20190102'],
            ['carol-gooddata', 'test', '20190101'],
            ['derek-lateisok', 'control', '20190110'],
        ],
        [
            "client_id",
            "branch",
            "enrollment_date",
        ],
    )

    ex_d = {'a-stub': 'fake-branch-lifes-too-short'}
    data_source = spark.createDataFrame(
        [
            # bob-badtiming only has data before/after analysis window
            # but missed by `filter_data_source_for_analysis_window`
            ['bob-badtiming', '20190102', ex_d, 1],
            ['bob-badtiming', '20190106', ex_d, 2],
            # carol-gooddata has data on two days (including a dupe day)
            ['carol-gooddata', '20190102', ex_d, 3],
            ['carol-gooddata', '20190102', ex_d, 2],
            ['carol-gooddata', '20190104', ex_d, 6],
            # derek-lateisok has data before and during the analysis window
            ['derek-lateisok', '20190110', ex_d, 1000],
            ['derek-lateisok', '20190111', ex_d, 1],
            # TODO: exercise the last condition on the join
        ],
        [
            "client_id",
            "submission_date_s3",
            "experiments",
            "some_value",
        ],
    )

    res = exp.get_per_client_data(
        enrollments,
        data_source,
        [
            F.coalesce(F.sum(data_source.some_value), F.lit(0)).alias('some_value'),
        ],
        '20190114',
        1,
        3,
        keep_client_id=True
    )

    # Check that the dataframe has the correct number of rows
    assert res.count() == enrollments.count()

    # Check that dataless enrollments are handled correctly
    annie_nodata = res.filter(res.client_id == 'annie-nodata')
    assert annie_nodata.count() == 1
    assert annie_nodata.first()['some_value'] == 0

    # Check that early and late data were ignored
    # i.e. check the join, not just _filter_data_source_for_analysis_window
    bob_badtiming = res.filter(res.client_id == 'bob-badtiming')
    assert bob_badtiming.count() == 1
    assert bob_badtiming.first()['some_value'] == 0
    # Check that _filter_data_source_for_analysis_window didn't do the
    # heavy lifting above
    fds = exp.filter_data_source_for_analysis_window(data_source, '20190114', 1, 3)
    assert fds.filter(
        fds.client_id == 'bob-badtiming'
    ).select(
        F.sum(fds.some_value).alias('agg_val')
    ).first()['agg_val'] == 3

    # Check that relevant data was included appropriately
    carol_gooddata = res.filter(res.client_id == 'carol-gooddata')
    assert carol_gooddata.count() == 1
    assert carol_gooddata.first()['some_value'] == 11

    derek_lateisok = res.filter(res.client_id == 'derek-lateisok')
    assert derek_lateisok.count() == 1
    assert derek_lateisok.first()['some_value'] == 1

    # Check that it still works for `data_source`s without an experiments map
    res2 = exp.get_per_client_data(
        enrollments,
        data_source.drop('experiments'),
        [
            F.coalesce(F.sum(data_source.some_value), F.lit(0)).alias('some_value'),
        ],
        '20190114',
        1,
        3,
        keep_client_id=True
    )

    assert res2.count() == enrollments.count()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pyspark.sql.functions as F
import pytest

from mozanalysis.utils import dedupe_columns, all_, any_, add_days


def test_dedupe_columns(spark):
    # NOTE: We are limited in the amount of testing we can do here b/c things
    # like `sorted([F.col(...)])` and checking for equality fail b/c of the
    # overridden magic methods on `Column` types.
    assert len(dedupe_columns(["a", "b", "a"])) == 2
    assert len(dedupe_columns(["a", "b", F.col("a")])) == 2
    assert len(dedupe_columns([F.col("a"), F.col("b"), F.col("a")])) == 2
    assert len(dedupe_columns([F.expr("a"), F.col("b"), F.col("a")])) == 2


def test_all_(spark):
    df = spark.createDataFrame(
        data=[
            [True, False, False],
            [True, False, True],
            [True, False, False],
            [True, False, True],
        ],
        schema=["all_true", "all_false", "mixed"],
    )
    pres = df.select(
        all_([df.all_true, df.all_true, df.all_true]).alias("true_1"),
        all_([df.all_true, df.all_false]).alias("false_2"),
        all_([df.all_false, df.all_false]).alias("false_3"),
        all_([df.mixed, df.all_false]).alias("false_4"),
        all_([df.mixed, df.all_true]).alias("mixed_5"),
    ).toPandas()
    assert pres.shape == (4, 5)
    assert not pres.isnull().any().any()
    assert pres["true_1"].all()
    assert not pres["false_2"].any()
    assert not pres["false_3"].any()
    assert not pres["false_4"].any()
    assert not pres["mixed_5"][::2].any()
    assert pres["mixed_5"][1::2].all()

    # Check this workaround is still necessary:
    with pytest.raises(ValueError):
        all([df.all_true, df.all_true])


def test_any_(spark):
    df = spark.createDataFrame(
        data=[
            [True, False, False],
            [True, False, True],
            [True, False, False],
            [True, False, True],
        ],
        schema=["all_true", "all_false", "mixed"],
    )
    pres = df.select(
        any_([df.all_true, df.all_true, df.all_true]).alias("true_1"),
        any_([df.all_true, df.all_false]).alias("true_2"),
        any_([df.all_false, df.all_false]).alias("false_3"),
        any_([df.mixed, df.all_true]).alias("true_4"),
        any_([df.mixed, df.all_false]).alias("mixed_5"),
    ).toPandas()
    assert pres.shape == (4, 5)
    assert not pres.isnull().any().any()
    assert pres["true_1"].all()
    assert pres["true_2"].any()
    assert not pres["false_3"].any()
    assert pres["true_4"].any()
    assert not pres["mixed_5"][::2].any()
    assert pres["mixed_5"][1::2].all()

    # Check this workaround is still necessary:
    with pytest.raises(ValueError):
        any([df.all_true, df.all_true])


def test_add_days():
    assert add_days('20190101', 0) == '20190101'
    assert add_days('20190101', 1) == '20190102'
    assert add_days('20190101', -1) == '20181231'



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from pyspark.sql import functions as F

from mozanalysis.utils import add_days


class Experiment(object):
    """Get DataFrames of experiment data; store experiment metadata.

    The methods here query data in a way compatible with the following
    principles, which are important for experiment analysis:

    - The population of clients in each branch must have the same
        properties, aside from the intervention itself and its
        consequences; i.e. there must be no underlying bias in the
        branch populations.
    - We must measure the same thing for each client, to minimize the
        variance associated with our measurement.

    So that our analyses follow these abstract principles, we follow
    these rules:

    - Start with a list of all clients who enrolled.
    - We can filter this list of clients only based on information known
        to us at or before the time that they enrolled, because later
        information might be causally connected to the intervention.
    - For any given metric, every client gets a non-null value; we don't
        implicitly ignore anyone, even if they churned and stopped
        sending data.
    - Typically if an enrolled client no longer qualifies for enrollment,
        we'll still want to include their data in the analysis, unless
        we're explicitly using stats methods that handle censored data.
    - We define a "analysis window" with respect to clients'
        enrollment dates. Each metric only uses data collected inside
        this analysis window. We can only analyze data for a client
        if we have data covering their entire analysis window.


    Example usage:
        cd = spark.table('clients_daily')

        experiment = Experiment(
            experiment_slug='pref-flip-defaultoncookierestrictions-1506704',
            start_date='20181215',
            num_dates_enrollment=8
        )
        enrollments = experiment.get_enrollments(spark)

        res = experiment.get_per_client_data(
            enrollments,
            cd,
            [
                F.sum(F.coalesce(
                    cd.active_hours_sum, F.lit(0)
                )).alias('active_hours'),
                F.sum(F.coalesce(
                    cd.scalar_parent_browser_engagement_total_uri_count_sum,
                    F.lit(0)
                )).alias('uri_count'),
            ],
            last_date_full_data='20190107',
            analysis_start_days=0,
            analysis_length_days=7
        )

        # Pull data into a pandas df, ready for running stats
        pres = res.toPandas()

    Args:
        experiment_slug (str): Name of the study, used to identify
            the enrollment events specific to this study.
        start_date (str): e.g. '20190101'. First date on which enrollment
            events were received.
        num_dates_enrollment (int, optional): Only include this many dates
            of enrollments. If `None` then use the maximum number of dates
            as determined by the metric's analysis window and
            `last_date_full_data`. Typically `7n+1`, e.g. `8`. The
            factor '7' removes weekly seasonality, and the `+1` accounts
            for the fact that enrollment typically starts a few hours
            before UTC midnight.
        addon_version (str, optional): The version of the experiment addon.
            Some addon experiment slugs get reused - in those cases we need
            to filter on the addon version also.

    Attributes:
        experiment_slug (str): Name of the study, used to identify
            the enrollment events specific to this study.
        start_date (str): e.g. '20190101'. First date on which enrollment
            events were received.
        num_dates_enrollment (int, optional): Only include this many days
            of enrollments. If `None` then use the maximum number of days
            as determined by the metric's analysis window and
            `last_date_full_data`. Typically `7n+1`, e.g. `8`. The
            factor '7' removes weekly seasonality, and the `+1` accounts
            for the fact that enrollment typically starts a few hours
            before UTC midnight.
        addon_version (str, optional): The version of the experiment addon.
            Some addon experiment slugs get reused - in those cases we need
            to filter on the addon version also.
    """
    def __init__(
        self, experiment_slug, start_date, num_dates_enrollment=None,
        addon_version=None
    ):
        # Let's be conservative about storing state - it doesn't belong
        # in this class. Treat these attributes as immutable.
        # These attributes are stored because it would be a PITA not to
        # store them:
        #   - they are required by both `get_enrollments()` and
        #       `get_per_client_data()`
        #   - if you were accidentally inconsistent with their values
        #       then you would hit trouble quickly!
        self.experiment_slug = experiment_slug
        self.start_date = start_date
        self.num_dates_enrollment = num_dates_enrollment
        self.addon_version = addon_version

    def get_enrollments(
        self, spark, study_type='pref_flip', end_date=None, debug_dupes=False
    ):
        """Return a DataFrame of enrolled clients.

        This works for pref-flip and addon studies.

        The underlying queries are different for pref-flip vs addon
        studies, because as of 2019/04/02, branch information isn't
        reliably available in the `events` table for addon experiments:
        branch may be NULL for all enrollments. The enrollment
        information for them is most reliably available in
        `telemetry_shield_study_parquet`. Once this issue is resolved,
        we will probably start using normandy events for all desktop
        studies.
        Ref: https://bugzilla.mozilla.org/show_bug.cgi?id=1536644

        Args:
            spark: The spark context.
            study_type (str): One of the following strings:
                - 'pref_flip'
                - 'addon'
            end_date (str, optional): Ignore enrollments after this
                date: for faster queries on stale experiments. If you
                set `num_dates_enrollment` then do not set this; at best
                it would be redundant, at worst it's contradictory.

        Returns:
            A Spark DataFrame of enrollment data. One row per
            enrollment. Columns:
                - client_id (str)
                - enrollment_date (str): e.g. '20190329'
                - branch (str)
        """
        if study_type == 'pref_flip':
            enrollments = self._get_enrollments_view_normandy(spark)

        elif study_type == 'addon':
            enrollments = self._get_enrollments_view_addon(spark, self.addon_version)

        # elif study_type == 'glean':
        #     raise NotImplementedError

        else:
            raise ValueError("Unrecognized study_type: {}".format(study_type))

        enrollments = enrollments.filter(
            enrollments.enrollment_date >= self.start_date
        ).filter(
            enrollments.experiment_slug == self.experiment_slug
        )

        if self.num_dates_enrollment is not None:
            if end_date is not None:
                raise ValueError(
                    "Don't specify both 'end_date' and "
                    "'num_dates_enrollment'; you might contradict yourself."
                )
            enrollments = enrollments.filter(
                enrollments.enrollment_date <= self._get_scheduled_max_enrollment_date()
            )
        elif end_date is not None:
            enrollments = enrollments.filter(
                enrollments.enrollment_date <= end_date
            )

        # Deduplicate enrollment events. Optionally keep track of what
        # had to be deduplicated. Deduplicating a client who enrolls in
        # multiple branches is left as an exercise to the reader :|
        enrollments = enrollments.groupBy(
            enrollments.client_id, enrollments.branch
        ).agg(*(
            [F.min(enrollments.enrollment_date).alias('enrollment_date')]
            + (
                [F.count(enrollments.enrollment_date).alias('num_events')]
                if debug_dupes else []
            )
        ))

        enrollments.cache()

        return enrollments

    def get_per_client_data(
        self, enrollments, data_source, metric_list, last_date_full_data,
        analysis_start_days, analysis_length_days, keep_client_id=False
    ):
        """Return a DataFrame containing per-client metric values.

        Args:
            enrollments: A spark DataFrame of enrollments, like the one
                returned by `self.get_enrollments()`.
            data_source: A spark DataFrame containing the data needed to
                calculate the metrics. Could be `main_summary` or
                `clients_daily`. _Don't_ use `experiments`; as of 2019/04/02
                it drops data collected after people self-unenroll, so
                unenrolling users will appear to churn. Must have at least
                the following columns:
                    - client_id (str)
                    - submission_date_s3 (str)
                    - data columns referred to in `metric_list`
                Ideally also has:
                    - experiments (map): At present this is used to exclude
                        pre-enrollment ping data collected on enrollment
                        day. Once it or its successor reliably tags data
                        from all enrolled users, even post-unenroll, we'll
                        also join on it to exclude data from duplicate
                        `client_id`s that are not enrolled in the same
                        branch.
            metric_list: A list of columns that aggregate and compute
                metrics over data grouped by `(client_id, branch)`, e.g.
                ```
                    [F.coalesce(F.sum(
                        data_source.metric_name
                    ), F.lit(0)).alias('metric_name')]
                ```
            last_date_full_data (str): The most recent date for which we
                have complete data, e.g. '20190322'. If you want to ignore
                all data collected after a certain date (e.g. when the
                experiment recipe was deactivated), then do that here.
            analysis_start_days (int): the start of the analysis window,
                measured in 'days since the client enrolled'. We ignore data
                collected outside this analysis window.
            analysis_length_days (int): the length of the analysis window,
                measured in days.
            keep_client_id (bool): Whether to return a `client_id` column.
                Defaults to False to reduce memory usage of the results.

        Returns:
            A spark DataFrame of experiment data. One row per `client_id`.
            One or two metadata columns, then one column per metric in
            `metric_list`. Then one column per sanity-check metric.
            Columns:
                - client_id (str, optional): Not necessary for
                    "happy path" analyses.
                - branch (str): The client's branch
                - [metric 1]: The client's value for the first metric in
                    `metric_list`.
                - ...
                - [metric n]: The client's value for the nth (final)
                    metric in `metric_list`.
                - [sanity check 1]: The client's value for the first
                    sanity check metric.
                - ...
                - [sanity check n]: The client's value for the last
                    sanity check metric.

            This format - the schema plus there being one row per
            enrolled client, regardless of whether the client has data
            in `data_source` - was agreed upon by the DS team, and is the
            standard format for queried experimental data.
        """
        for col in ['client_id', 'submission_date_s3']:
            if col not in data_source.columns:
                raise ValueError("Column '{}' missing from 'data_source'".format(col))

        req_dates_of_data = analysis_start_days + analysis_length_days

        enrollments = self.filter_enrollments_for_analysis_window(
            enrollments, last_date_full_data, req_dates_of_data
        )

        data_source = self.filter_data_source_for_analysis_window(
            data_source, last_date_full_data, analysis_start_days,
            analysis_length_days
        )

        join_on = [
            # TODO perf: would it be faster if we enforce a join on sample_id?
            enrollments.client_id == data_source.client_id,

            # TODO accuracy: once we can rely on
            #   `data_source.experiments[self.experiment_slug]`
            # existing even after unenrollment, we could start joining on
            # branch to reduce problems associated with split client_ids:
            # enrollments.branch == data_source.experiments[self.experiment_slug]

            # Do a quick pass aiming to efficiently filter out lots of rows:
            enrollments.enrollment_date <= data_source.submission_date_s3,

            # Now do a more thorough pass filtering out irrelevant data:
            # TODO perf: what is a more efficient way to do this?
            (
                (
                    F.unix_timestamp(data_source.submission_date_s3, 'yyyyMMdd')
                    - F.unix_timestamp(enrollments.enrollment_date, 'yyyyMMdd')
                ) / (24 * 60 * 60)
            ).between(
                analysis_start_days,
                analysis_start_days + analysis_length_days - 1
            ),
        ]

        if 'experiments' in data_source.columns:
            # Try to filter data from day of enrollment before time of enrollment.
            # If the client enrolled and unenrolled on the same day then this
            # will also filter out that day's post unenrollment data but that's
            # probably the smallest and most innocuous evil on the menu.
            join_on.append(
                (enrollments.enrollment_date != data_source.submission_date_s3)
                | (~F.isnull(data_source.experiments[self.experiment_slug]))
            ),

        sanity_metrics = self._get_telemetry_sanity_check_metrics(
            enrollments, data_source
        )

        res = enrollments.join(
            data_source,
            join_on,
            'left'
        ).groupBy(
            enrollments.client_id, enrollments.branch
        ).agg(
            *(metric_list + sanity_metrics)
        )
        if keep_client_id:
            return res
        else:
            return res.drop(enrollments.client_id)

    @staticmethod
    def _get_enrollments_view_normandy(spark):
        """Return a DataFrame of all normandy enrollment events.

        Filter the `events` table to enrollment events and transform it
        into the standard enrollments schema.

        Args:
            spark: The spark context.
        """
        events = spark.table('events')

        return events.filter(
            events.event_category == 'normandy'
        ).filter(
            events.event_method == 'enroll'
        ).select(
            events.client_id,
            events.event_string_value.alias('experiment_slug'),
            events.event_map_values.branch.alias('branch'),
            events.submission_date_s3.alias('enrollment_date'),
        )

    @staticmethod
    def _get_enrollments_view_addon(spark, addon_version=None):
        """Return a DataFrame of all addon study enrollment events.

        Filter the `telemetry_shield_study_parquet` to enrollment events
        and transform it into the standard enrollments schema.

        Args:
            spark: The spark context.
            addon_version (str, optional): The version of the experiment
                addon. Some addon experiment slugs get reused - in those
                cases we need to filter on the addon version also.
        """
        tssp = spark.table('telemetry_shield_study_parquet')

        if addon_version is not None:
            # It's a little messy that we filter `addon_version` here
            # and `study_name` elsewhere, but this seemed the least
            # hacky solution
            tssp = tssp.filter(
                tssp.payload.addon_version == addon_version
            )

        return tssp.filter(
            tssp.payload.data.study_state == 'enter'
        ).select(
            tssp.client_id,
            tssp.payload.study_name.alias('experiment_slug'),
            tssp.payload.branch.alias('branch'),
            tssp.submission.alias('enrollment_date'),
        )

    def _get_scheduled_max_enrollment_date(self):
        """Return the last enrollment date, according to the plan."""
        assert self.num_dates_enrollment is not None

        return add_days(self.start_date, self.num_dates_enrollment - 1)

    def _get_last_enrollment_date(self, last_date_full_data, req_dates_of_data):
        """Return the date of the final used enrollment.

        We need `req_dates_of_data` days of post-enrollment data per client.
        This and `last_date_full_data` put constraints on the enrollment
        period. This method checks these constraints are feasible, and
        compatible with any manually supplied enrollment period.

        If `self.num_dates_enrollment` is `None`, then there is potential
        for the final date to be surprising, so we print it.

        Args:
            last_date_full_data (str): The most recent date for which we
                have complete data, e.g. '20190322'. If you want to ignore
                all data collected after a certain date (e.g. when the
                experiment recipe was deactivated), then do that here.
            req_dates_of_data (int): The minimum number of dates of
                post-enrollment data required to have data for the client
                for the entire analysis window.
        """
        last_enrollment_with_data = add_days(
            last_date_full_data, -(req_dates_of_data - 1)
        )

        if self.num_dates_enrollment is None:
            if last_enrollment_with_data < self.start_date:
                raise ValueError("No users have a complete analysis window")

            print("Taking enrollments between {} and {}".format(
                self.start_date, last_enrollment_with_data
            ))

            return last_enrollment_with_data

        else:
            intended_last_enrollment = self._get_scheduled_max_enrollment_date()

            if last_enrollment_with_data < intended_last_enrollment:
                raise ValueError(
                    "You said you wanted {} dates of enrollment, ".format(
                        self.num_dates_enrollment
                    ) + "but your analysis window of {} days won't have ".format(
                        req_dates_of_data
                    ) + "complete data until we have the data for {}.".format(
                        add_days(intended_last_enrollment, req_dates_of_data - 1)
                    )
                )

            return intended_last_enrollment

    def _get_last_data_date(self, last_date_full_data, req_dates_of_data):
        """Return the date of the final used datum."""
        last_enrollment_date = self._get_last_enrollment_date(
            last_date_full_data, req_dates_of_data
        )

        last_required_data_date = add_days(
            last_enrollment_date, req_dates_of_data - 1
        )

        # `_get_last_enrollment_date` should have checked for this
        assert last_required_data_date <= last_date_full_data

        return last_required_data_date

    def filter_enrollments_for_analysis_window(
        self, enrollments, last_date_full_data, req_dates_of_data
    ):
        """Return the enrollments, filtered to the relevant dates."""
        return enrollments.filter(
            # Ignore clients without a complete analysis window
            enrollments.enrollment_date <= self._get_last_enrollment_date(
                last_date_full_data, req_dates_of_data
            )
        )

    def filter_data_source_for_analysis_window(
        self, data_source, last_date_full_data, analysis_start_days,
        analysis_length_days
    ):
        """Return `data_source`, filtered to the relevant dates.

        This should not affect the results - it should just speed things
        up.
        """
        return data_source.filter(
            # Ignore data before the analysis window of the first enrollment
            data_source.submission_date_s3 >= add_days(
                self.start_date, analysis_start_days
            )
        ).filter(
            # Ignore data after the analysis window of the last enrollment,
            # and data after the specified `last_date_full_data`
            data_source.submission_date_s3 <= self._get_last_data_date(
                last_date_full_data, analysis_start_days + analysis_length_days
            )
        )

    def _get_telemetry_sanity_check_metrics(self, enrollments, data_source):
        """Return aggregations that check for problems with a client."""

        # TODO: Once we know what form the metrics library will take,
        # we should move the below metric definitions and documentation
        # into it.

        if dict(data_source.dtypes).get('experiments') != 'map<string,string>':
            # Not all tables have an experiments map - can't make these checks.
            return []

        return [

            # Check to see whether the client_id is also enrolled in other branches
            # E.g. indicates cloned profiles. Fraction of such users should be
            # small, and similar between branches.
            F.max(F.coalesce((
                data_source.experiments[self.experiment_slug] != enrollments.branch
            ).astype('int'), F.lit(0))).alias('has_contradictory_branch'),

            # Check to see whether the client_id was sending data in the analysis
            # window that wasn't tagged as being part of the experiment. Indicates
            # either a client_id clash, or the client unenrolling. Fraction of such
            # users should be small, and similar between branches.
            F.max(F.coalesce((
                ~F.isnull(data_source.experiments)
                & F.isnull(data_source.experiments[self.experiment_slug])
            ).astype('int'), F.lit(0))).alias('has_non_enrolled_data'),
        ]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import datetime
from functools import reduce

import pyspark.sql.functions as F
from pyspark.sql.column import Column


def dedupe_columns(columns):
    """
    Given a list of columns, returns a list with duplicates removed.

    These can be either from `F.col` or `F.expr`.

    NOTE: Because of the underlying way `Column`s are coded using Python
    magic methods, the usual methods for comparing or deduping don't work
    here. They aren't hashable so we can't use sets and they aren't
    comparable by normal methods. So we fall back to comparing the underlying
    Java object ids instead.

    """
    d = {}

    for col in columns:
        if not isinstance(col, Column):
            col = F.col(col)
        d[col._jc] = col

    return list(d.values())


def all_(l):
    """Return the element-wise logical AND of `Column`s.

    Think of this as a vector-friendly version of the built-in function
    `all()`.

    Args:
        l: A list of `Column`s of booleans. Or, more generally,
            an iterable of items to reduce over.

    Returns:
        A `Column` of booleans representing the logical AND. Or,
            more generally, the result of the logical AND.
    """
    return reduce(lambda x, y: x & y, l)


def any_(l):
    """Return the element-wise logical OR of `Column`s.

    Think of this as a vector-friendly version of the built-in function
    `any()`.

    Args:
        l: A list of `Column`s of booleans. Or, more generally,
            an iterable of items to reduce over.

    Returns:
        A `Column` of booleans representing the logical OR. Or,
            more generally, the result of the logical OR.
    """
    return reduce(lambda x, y: x | y, l)


def add_days(date_string, n_days):
    """Add `n_days` days to a date string like '20190101'."""
    original_date = datetime.datetime.strptime(date_string, '%Y%m%d')
    new_date = original_date + datetime.timedelta(days=n_days)
    return datetime.datetime.strftime(new_date, '%Y%m%d')



from pyspark.sql import Row
from pyspark.sql import functions as F
from pytest import approx


def generate_histogram(spark):
    return spark.createDataFrame(
        [
            Row(client_id="a", my_histogram={"0": 5, "10": 10, "42": 1, "43": 0}),
            Row(client_id="b", my_histogram={"0": 10, "10": 0, "42": 38, "43": 5}),
        ]
    )


def test_histogram_udfs(spark):
    from mozanalysis.udf import histogram_mean, histogram_count, histogram_sum

    df = (
        generate_histogram(spark)
        .withColumn("mean", histogram_mean("my_histogram"))
        .withColumn("count", histogram_count("my_histogram"))
        .withColumn("sum", histogram_sum("my_histogram"))
        .toPandas()
        .set_index("client_id")
    )
    assert df.loc["a", "count"] == 16
    assert df.loc["b", "count"] == 53
    assert df.loc["a", "sum"] == 142
    assert df.loc["b", "sum"] == 38 * 42 + 5 * 43
    assert df.loc["a", "mean"] == float(df.loc["a", "sum"]) / df.loc["a", "count"]
    assert df.loc["b", "mean"] == float(df.loc["b", "sum"]) / df.loc["b", "count"]


def test_histogram_quantiles(spark):
    from mozanalysis.udf import generate_quantile_udf

    sdf = generate_histogram(spark)
    quantile_udf = generate_quantile_udf(
        sdf=sdf,
        grouping_fields=["client_id"],
        bucket_field="bucket",
        count_field="count",
        quantiles=[0.5, 0.95],
    )

    df = (
        sdf
        .select(
            sdf.client_id,
            F.explode(sdf.my_histogram).alias("bucket", "count")
        )
        .groupBy(sdf.client_id)
        .apply(quantile_udf)
        .toPandas()
        .set_index(["client_id", "quantile"])
    )

    assert df.loc[("a", 0.5), "value"] == 10
    assert df.loc[("a", 0.95), "value"] == 42
    assert df.loc[("b", 0.5), "value"] == 42
    assert df.loc[("b", 0.95), "value"] == 43


def test_histogram_threshold(spark):
    from mozanalysis.udf import generate_threshold_udf

    sdf = generate_histogram(spark)
    threshold_udf = generate_threshold_udf(
        sdf=sdf,
        grouping_fields=["client_id"],
        bucket_field="bucket",
        count_field="count",
        thresholds=[0, 43],
    )

    df = (
        sdf
        .select(
            sdf.client_id,
            F.explode(sdf.my_histogram).alias("bucket", "count")
        )
        .groupBy(sdf.client_id)
        .apply(threshold_udf)
        .toPandas()
        .set_index(["client_id", "threshold"])
    )

    def check(client, threshold, fraction):
        value = df.loc[(client, threshold), "fraction_exceeding"]
        assert value == approx(fraction)

    check("a", 0, 1)
    check("a", 43, 0)
    check("b", 0, 1)
    check("b", 43, 5.0 / 53)
